/*
 *  nautilus-gdu-window.c
 *
 *  Copyright (C) 2008-2009 Red Hat, Inc.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Author: Tomas Bzatek <tbzatek@redhat.com>
 *
 */

#ifdef HAVE_CONFIG_H
 #include <config.h> /* for GETTEXT_PACKAGE */
#endif

#include "nautilus-gdu.h"
#include "nautilus-gdu-window.h"

#include <glib/gi18n-lib.h>
#include <gio/gio.h>

#include <gtk/gtkicontheme.h>
#include <gtk/gtkwidget.h>
#include <gtk/gtkmain.h>

#include <gdu/gdu.h>
#include <gdu-gtk/gdu-gtk.h>

#include <polkit-gnome/polkit-gnome.h>




typedef struct {
  GtkDialog *dialog;
  GtkWidget *close_button;
  GtkWidget *icon_image;
  GtkWidget *name_label;
  GtkWidget *details1_label;
  GtkWidget *details2_label;
  GtkWidget *details3_label;
  GtkWidget *mount_warning;
  GtkWidget *label_entry;
  GtkWidget *part_type_combo_box;
  GtkWidget *part_flag_boot_check_button;
  GtkWidget *part_flag_required_check_button;
  GtkWidget *encrypted_check_button;
  GtkWidget *take_ownership_of_fs_check_button;
  GtkWidget *progress_bar;
  GtkWidget *progress_bar_box;
  GtkWidget *controls_box;
  GtkWidget *partition_label;

  GduPresentable *presentable;
  GduPool *pool;

  gboolean job_running;
  gboolean job_cancelled;

  PolKitAction *pk_unmount_action;
  PolKitGnomeAction *unmount_action;
  PolKitAction *pk_format_action;
  PolKitGnomeAction *format_action;
  PolKitAction *pk_part_modify_action;
  PolKitGnomeAction *part_modify_action;
} FormatDialogPrivate;

typedef struct {
  FormatDialogPrivate *priv;
  char *encrypt_passphrase;
  gboolean save_in_keyring;
  gboolean save_in_keyring_session;
  char *fslabel;
  char *fstype;
  GduDevice *device;
  gboolean take_ownership;
  char *secure_erase;
  const char *recommended_part_type;

  guint job_progress_pulse_timer_id;
  gulong job_handler_id;
  gulong pk_unmount_action_handler_id;
  gulong pk_unmount_auth_end_handler_id;
  gulong pk_format_action_handler_id;
  gulong pk_format_auth_end_handler_id;
  gulong pk_part_modify_action_handler_id;
  gulong pk_part_modify_auth_end_handler_id;
} FormatProcessData;




/*  taken from palimpsest/gdu-section-partition.c  */
static gboolean
has_flag (char **flags, const char *flag)
{
  int n = 0;
  while (flags != NULL && flags[n] != NULL) {
    if (strcmp (flags[n], flag) == 0)
      return TRUE;
    n++;
  }
  return FALSE;
}

/*  TODO: this is quite dangerous, but needed  */
static const char *
get_part_type_for_fs (const char *fs_type)
{
  if (strcasecmp (fs_type, "vfat") == 0)
    return "0x0c";
  else
  if (strcasecmp (fs_type, "empty") == 0 || strlen (fs_type) == 0)
    return "0x0";
  else
  if (strcasecmp (fs_type, "ntfs") == 0)
    return "0x07";
  else
  if (strcasecmp (fs_type, "swap") == 0)
    return "0x82";
  else
    return "0x83";   /* default to "linux"  */
}

static gboolean
mbr_part_is_linux_type (const char *part_type)
{
  return (strcasecmp (part_type, "0x83") == 0 || strcasecmp (part_type, "0x8e") == 0 ||
          strcasecmp (part_type, "0xfd") == 0);
}


/*  taken from palimpsest/gdu-section-unrecognized.c  */
static void
type_combo_box_changed (GtkWidget *combo_box, FormatDialogPrivate *priv)
{
  char *fstype;
  GduKnownFilesystem *kfs;
  gboolean label_entry_sensitive;
//  gboolean can_erase;
  gboolean have_owners;
  int max_label_len;

  fstype = NULL;

  label_entry_sensitive = FALSE;
//  can_erase = FALSE;
  max_label_len = 0;
  have_owners = FALSE;

  fstype = gdu_util_fstype_combo_box_get_selected (combo_box);
  g_print ("type_combo_box_changed: fstype = '%s'\n", fstype);
  if (fstype != NULL) {
    kfs = gdu_pool_get_known_filesystem_by_id (priv->pool, fstype);
    if (kfs != NULL) {
      g_print ("         kfs ID = '%s', name = %s\n", gdu_known_filesystem_get_id (kfs), gdu_known_filesystem_get_name (kfs));
      max_label_len = gdu_known_filesystem_get_max_label_len (kfs);
      have_owners = gdu_known_filesystem_get_supports_unix_owners (kfs);
      g_object_unref (kfs);
    }
//    can_erase = TRUE;
  }

  if (max_label_len > 0)
    label_entry_sensitive = TRUE;

  gtk_entry_set_max_length (GTK_ENTRY (priv->label_entry), max_label_len);
  gtk_widget_set_sensitive (priv->label_entry, label_entry_sensitive);
//  polkit_gnome_action_set_sensitive (priv->erase_action, can_erase);

  if (have_owners)
    gtk_widget_show (priv->take_ownership_of_fs_check_button);
  else
    gtk_widget_hide (priv->take_ownership_of_fs_check_button);

  g_free (fstype);
}


static void
update_ui_mounted (FormatDialogPrivate *priv)
{
  GduDevice *device;

  g_return_if_fail (GDU_IS_PRESENTABLE (priv->presentable));

  device = gdu_presentable_get_device (priv->presentable);

  if (device && gdu_device_is_mounted (device))
    gtk_widget_show_all (priv->mount_warning);
  else
    gtk_widget_hide_all (priv->mount_warning);

  if (device != NULL)
    g_object_unref (device);
}

static gboolean
job_progress_pulse_timeout_handler (gpointer user_data)
{
  FormatProcessData *data = user_data;

  g_return_val_if_fail (data != NULL, TRUE);

  gtk_progress_bar_pulse (GTK_PROGRESS_BAR (data->priv->progress_bar));
  return TRUE;
}

static void
presentable_job_changed (GduPresentable *presentable, gpointer user_data)
{
  FormatProcessData *data = user_data;
  char *s;
  char *job_description;
  char *task_description;
  double percentage;

  g_return_if_fail (data != NULL);

  if (data->device != NULL && gdu_device_job_in_progress (data->device)) {
          job_description = gdu_get_job_description (gdu_device_job_get_id (data->device));
          task_description = gdu_get_task_description (gdu_device_job_get_cur_task_id (data->device));

          s = g_strdup_printf (_("%s (task %d of %d)"),
                               (task_description && strlen (task_description) > 0) ? task_description : job_description,
                               gdu_device_job_get_cur_task (data->device) + 1,
                               gdu_device_job_get_num_tasks (data->device));
          gtk_progress_bar_set_text (GTK_PROGRESS_BAR (data->priv->progress_bar), s);
          g_free (s);

          percentage = gdu_device_job_get_cur_task_percentage (data->device);
          if (percentage < 0) {
                  gtk_progress_bar_set_pulse_step (GTK_PROGRESS_BAR (data->priv->progress_bar), 2.0 / 50);
                  gtk_progress_bar_pulse (GTK_PROGRESS_BAR (data->priv->progress_bar));
                  if (data->job_progress_pulse_timer_id == 0) {
                    data->job_progress_pulse_timer_id = g_timeout_add (
                                  1000 / 50,
                                  job_progress_pulse_timeout_handler,
                                  data);
                  }
          } else {
                  gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (data->priv->progress_bar),
                                                 percentage / 100.0);
                  if (data->job_progress_pulse_timer_id > 0) {
                          g_source_remove (data->job_progress_pulse_timer_id);
                          data->job_progress_pulse_timer_id = 0;
                  }
          }

          g_free (job_description);
          g_free (task_description);

  } else {
          if (data->job_progress_pulse_timer_id > 0) {
                  g_source_remove (data->job_progress_pulse_timer_id);
                  data->job_progress_pulse_timer_id = 0;
          }
  }
}

static void unmount_auth_end_callback (PolKitGnomeAction *action, gboolean gained_privilege, gpointer user_data);
static void unmount_action_callback (GtkAction *action, gpointer user_data);
static void format_auth_end_callback (PolKitGnomeAction *action, gboolean gained_privilege, gpointer user_data);
static void format_action_callback (GtkAction *action, gpointer user_data);
static void part_modify_auth_end_callback (PolKitGnomeAction *action, gboolean gained_privilege, gpointer user_data);
static void part_modify_action_callback (GtkAction *action, gpointer user_data);

static void
update_ui_progress (FormatDialogPrivate *priv,
                    FormatProcessData *data,
                    gboolean working)
{
  g_return_if_fail (priv != NULL);

//  g_print ("update_ui_progress\n");

  priv->job_running = working;

  if (working) {
    gtk_progress_bar_set_text (GTK_PROGRESS_BAR (priv->progress_bar), NULL);
    gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (priv->progress_bar), 0.0);
    gtk_button_set_label (GTK_BUTTON (priv->close_button), GTK_STOCK_STOP);
    gtk_widget_show_all (priv->progress_bar_box);
    if (data) {
      data->job_handler_id = g_signal_connect (priv->presentable, "job-changed", G_CALLBACK (presentable_job_changed), data);
      data->pk_unmount_auth_end_handler_id = g_signal_connect (priv->unmount_action, "auth-end", G_CALLBACK (unmount_auth_end_callback), data);
      data->pk_unmount_action_handler_id = g_signal_connect (priv->unmount_action, "activate", G_CALLBACK (unmount_action_callback), data);
      data->pk_format_auth_end_handler_id = g_signal_connect (priv->format_action, "auth-end", G_CALLBACK (format_auth_end_callback), data);
      data->pk_format_action_handler_id = g_signal_connect (priv->format_action, "activate", G_CALLBACK (format_action_callback), data);
      data->pk_part_modify_auth_end_handler_id = g_signal_connect (priv->part_modify_action, "auth-end", G_CALLBACK (part_modify_auth_end_callback), data);
      data->pk_part_modify_action_handler_id = g_signal_connect (priv->part_modify_action, "activate", G_CALLBACK (part_modify_action_callback), data);
    }
  }
  else
  {
    if (data) {
      if (data->job_handler_id)
        g_signal_handler_disconnect (priv->presentable, data->job_handler_id);
      if (data->pk_unmount_auth_end_handler_id)
        g_signal_handler_disconnect (priv->unmount_action, data->pk_unmount_auth_end_handler_id);
      if (data->pk_unmount_action_handler_id)
        g_signal_handler_disconnect (priv->unmount_action, data->pk_unmount_action_handler_id);
      if (data->pk_format_auth_end_handler_id)
        g_signal_handler_disconnect (priv->format_action, data->pk_format_auth_end_handler_id);
      if (data->pk_format_action_handler_id)
        g_signal_handler_disconnect (priv->format_action, data->pk_format_action_handler_id);
      if (data->pk_part_modify_auth_end_handler_id)
        g_signal_handler_disconnect (priv->part_modify_action, data->pk_part_modify_auth_end_handler_id);
      if (data->pk_part_modify_action_handler_id)
        g_signal_handler_disconnect (priv->part_modify_action, data->pk_part_modify_action_handler_id);
      if (data->job_progress_pulse_timer_id > 0) {
        g_source_remove (data->job_progress_pulse_timer_id);
        data->job_progress_pulse_timer_id = 0;
      }
    }
    gtk_widget_hide_all (priv->progress_bar_box);
    gtk_button_set_label (GTK_BUTTON (priv->close_button), GTK_STOCK_CLOSE);
  }
  gtk_widget_set_sensitive (priv->controls_box, ! working);
  gtk_widget_set_sensitive (priv->partition_label, ! working);
  gtk_dialog_set_response_sensitive (priv->dialog, GTK_RESPONSE_OK, ! working);
  gtk_dialog_set_response_sensitive (priv->dialog, GTK_RESPONSE_REJECT, ! working);
//  gtk_dialog_set_response_sensitive (priv->dialog, GTK_RESPONSE_CLOSE, ! working);
}


/*  most of the code here was stolen from palimpsest  */
static void
update_ui (FormatDialogPrivate *priv)
{
  char *details1;
  char *details2;
  char *details3;
  char *s;
  char *name;
  char *icon_name;
  GdkPixbuf *pixbuf;
  GduDevice *device;
  const char *usage;
  const char *type;
  const char *device_file;
  char *strsize;
  GduPresentable *toplevel_presentable;
  GduDevice *toplevel_device = NULL;
  gboolean show_flag_boot;
  gboolean show_flag_required;
//  gboolean can_edit_part_label;
  const char *scheme;
  char **flags;


  device = gdu_presentable_get_device (priv->presentable);

  toplevel_presentable = gdu_presentable_get_toplevel (priv->presentable);
  if (toplevel_presentable != NULL)
    toplevel_device = gdu_presentable_get_device (toplevel_presentable);

  icon_name = gdu_presentable_get_icon_name (priv->presentable);
  name = gdu_presentable_get_name (priv->presentable);

  /*  window title  */
  s = g_strdup_printf (_("Format %s"), name);
  gtk_window_set_title (GTK_WINDOW (priv->dialog), s);
  g_free (s);


  /*  icon  */
  pixbuf = NULL;
  /*  we can also use gdu_util_get_pixbuf_for_presentable()  */
  if (icon_name != NULL) {
    pixbuf = gtk_icon_theme_load_icon (gtk_icon_theme_get_default (),
                                       icon_name,
                                       96,
                                       GTK_ICON_LOOKUP_GENERIC_FALLBACK,
                                       NULL);
  }
  gtk_image_set_from_pixbuf (GTK_IMAGE (priv->icon_image), pixbuf);
  g_object_unref (pixbuf);

  s = g_strdup_printf ("<span font_desc='18'><b>%s</b></span>", name);
  gtk_label_set_markup (GTK_LABEL (priv->name_label), s);
  g_free (s);

  /*  Is device mounted?  */
  update_ui_mounted (priv);


  usage = NULL;
  type = NULL;
  device_file = NULL;
  if (device != NULL) {
    usage = gdu_device_id_get_usage (device);
    type = gdu_device_id_get_type (device);
    device_file = gdu_device_get_device_file (device);
  }

  strsize = gdu_util_get_size_for_display (gdu_presentable_get_size (priv->presentable), FALSE);

  details1 = NULL;
  details2 = NULL;
  details3 = NULL;

  /*  Labels, header  */
  /*  TODO: clean unused code up  */
  if (GDU_IS_DRIVE (priv->presentable)) {
          details3 = g_strdup (device_file);
          if (GDU_IS_ACTIVATABLE_DRIVE (priv->presentable)) {
                  switch (gdu_activatable_drive_get_kind (GDU_ACTIVATABLE_DRIVE (priv->presentable))) {
                  case GDU_ACTIVATABLE_DRIVE_KIND_LINUX_MD:
                          details1 = g_strdup (_("Linux Software RAID"));
                          break;
                  default:
                          details1 = g_strdup (_("Activatable Drive"));
                          break;
                  }
          } else {
                  s = gdu_util_get_connection_for_display (
                          gdu_device_drive_get_connection_interface (device),
                          gdu_device_drive_get_connection_speed (device));
                  details1 = g_strdup_printf (_("Connected via %s"), s);
                  g_free (s);
          }

          if (device == NULL) {
                  /* TODO */
          } else {
                  if (gdu_device_is_removable (device)) {
                          if (gdu_device_is_partition_table (device)) {
                                  const char *scheme;
                                  scheme = gdu_device_partition_table_get_scheme (device);
                                  if (strcmp (scheme, "apm") == 0) {
                                          s = g_strdup (_("Apple Partition Map"));
                                  } else if (strcmp (scheme, "mbr") == 0) {
                                          s = g_strdup (_("Master Boot Record"));
                                  } else if (strcmp (scheme, "gpt") == 0) {
                                          s = g_strdup (_("GUID Partition Table"));
                                  } else {
                                          s = g_strdup_printf (_("Unknown Scheme: %s"), scheme);
                                  }

                                  details2 = g_strdup_printf (_("%s Partitioned Media (%s)"), strsize, s);

                                  g_free (s);
                          } else if (usage != NULL && strlen (usage) > 0) {
                                  details2 = g_strdup_printf (_("%s Unpartitioned Media"), strsize);
                          } else if (!gdu_device_is_media_available (device)) {
                                  details2 = g_strdup_printf (_("No Media Detected"));
                          } else {
                                  details2 = g_strdup_printf (_("Unrecognized"));
                          }
                  } else {
                          if (gdu_device_is_partition_table (device)) {
                                  const char *scheme;
                                  scheme = gdu_device_partition_table_get_scheme (device);
                                  if (strcmp (scheme, "apm") == 0) {
                                          s = g_strdup (_("Apple Partition Map"));
                                  } else if (strcmp (scheme, "mbr") == 0) {
                                          s = g_strdup (_("Master Boot Record"));
                                  } else if (strcmp (scheme, "gpt") == 0) {
                                          s = g_strdup (_("GUID Partition Table"));
                                  } else {
                                          s = g_strdup_printf (_("Unknown Scheme: %s"), scheme);
                                  }
                                  details2 = g_strdup_printf (_("Partitioned (%s)"), s);
                                  g_free (s);
                          } else if (usage != NULL && strlen (usage) > 0) {
                                  details2 = g_strdup_printf (_("Not Partitioned"));
                          } else if (!gdu_device_is_media_available (device)) {
                                  details2 = g_strdup_printf (_("No Media Detected"));
                          } else {
                                  details2 = g_strdup_printf (_("Unrecognized"));
                          }
                  }

                  if (gdu_device_is_read_only (device)) {
                          s = details3;
                          details3 = g_strconcat (details3, _(" (Read Only)"), NULL);
                          g_free (s);
                  }
          }
  } else if (GDU_IS_VOLUME (priv->presentable)) {
          details3 = g_strdup (device_file);

          if (strcmp (usage, "filesystem") == 0) {
                  char *fsname;
                  fsname = gdu_util_get_fstype_for_display (
                          gdu_device_id_get_type (device),
                          gdu_device_id_get_version (device),
                          TRUE);
                  details1 = g_strdup_printf (_("%s %s File System"), strsize, fsname);
                  g_free (fsname);
          } else if (strcmp (usage, "raid") == 0) {
                  char *fsname;
                  fsname = gdu_util_get_fstype_for_display (
                          gdu_device_id_get_type (device),
                          gdu_device_id_get_version (device),
                          TRUE);
                  details1 = g_strdup_printf (_("%s %s"), strsize, fsname);
                  g_free (fsname);
          } else if (strcmp (usage, "crypto") == 0) {
                  details1 = g_strdup_printf (_("%s Encrypted LUKS Device"), strsize);
          } else if (strcmp (usage, "other") == 0) {
                  if (strcmp (type, "swap") == 0) {
                          details1 = g_strdup_printf (_("%s Swap Space"), strsize);
                  } else {
                          details1 = g_strdup_printf (_("%s Data"), strsize);
                  }
          } else {
                  details1 = g_strdup_printf (_("%s Unrecognized"), strsize);
          }

          if (gdu_device_is_luks_cleartext (device)) {
                  details2 = g_strdup (_("Unlocked Encrypted LUKS Volume"));
          } else {
                  if (gdu_device_is_partition (device)) {
                          char *part_desc;
                          part_desc = gdu_util_get_desc_for_part_type (gdu_device_partition_get_scheme (device),
                                                                       gdu_device_partition_get_type (device));
                          details2 = g_strdup_printf (_("Partition %d (%s)"),
                                              gdu_device_partition_get_number (device), part_desc);
                          g_free (part_desc);
                  } else {
                          details2 = g_strdup (_("Not Partitioned"));
                  }
          }

          if (gdu_device_is_read_only (device)) {
                  s = details3;
                  details3 = g_strconcat (details3, _(" (Read Only)"), NULL);
                  g_free (s);
          }

          if (gdu_device_is_mounted (device)) {
                  s = details3;
                  details3 = g_strconcat (details3,
                                          _(" mounted at "),
                                          gdu_device_get_mount_path (device),
                                          NULL);
                  g_free (s);
          }
  } else if (GDU_IS_VOLUME_HOLE (priv->presentable)) {

          details1 = g_strdup_printf (_("%s Unallocated"), strsize);

          if (toplevel_device != NULL) {
                  details2 = g_strdup (gdu_device_get_device_file (toplevel_device));

                  if (gdu_device_is_read_only (toplevel_device)) {
                          s = details2;
                          details2 = g_strconcat (details2, _(" (Read Only)"), NULL);
                          g_free (s);
                  }
          }
  }

  g_print ("gdu_device_id_get_type (device) = '%s', gdu_device_partition_get_type (device) = '%s'\n",
           gdu_device_id_get_type (device), gdu_device_partition_get_type (device));

  g_free (icon_name);
  g_free (name);
  g_free (strsize);

  gtk_label_set_markup (GTK_LABEL (priv->details1_label), details1);
  gtk_label_set_markup (GTK_LABEL (priv->details2_label), details2);
  gtk_label_set_markup (GTK_LABEL (priv->details3_label), details3);
  g_free (details1);
  g_free (details2);
  g_free (details3);


  /*  partition type combo  */
  if (! priv->job_running) {
    gdu_util_fstype_combo_box_rebuild (priv->part_type_combo_box, priv->pool, NULL);
    gdu_util_fstype_combo_box_select (priv->part_type_combo_box, (type && strlen (type) > 0) ? type : "empty");
  }

  /*  encryption check button  */
  if (! gdu_pool_supports_luks_devices (priv->pool)) {
    gtk_widget_hide (priv->encrypted_check_button);
  }


  /*  TODO: is this really valid for a formatter?  -- disabled  */
  scheme = device ? gdu_device_partition_get_scheme (device) : NULL;
//  can_edit_part_label = FALSE;
  show_flag_boot = FALSE;
  show_flag_required = FALSE;

#if 0   /*  -- disabled  */
  if (strcmp (scheme, "mbr") == 0) {
//        can_edit_part_label = TRUE;
        show_flag_boot = TRUE;
  }

  if (strcmp (scheme, "gpt") == 0) {
//          can_edit_part_label = TRUE;
          show_flag_required = TRUE;
  }

  if (strcmp (scheme, "apm") == 0) {
//          can_edit_part_label = TRUE;
          show_flag_boot = TRUE;
  }
#endif

  if (show_flag_boot)
          gtk_widget_show (priv->part_flag_boot_check_button);
  else
          gtk_widget_hide (priv->part_flag_boot_check_button);

  if (show_flag_required)
          gtk_widget_show (priv->part_flag_required_check_button);
  else
          gtk_widget_hide (priv->part_flag_required_check_button);

  flags = gdu_device_partition_get_flags (device);
  if (! priv->job_running) {
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (priv->part_flag_boot_check_button),
                                  has_flag (flags, "boot"));
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (priv->part_flag_required_check_button),
                                  has_flag (flags, "required"));
  }

/*
  gtk_widget_set_sensitive (priv->label_entry, can_edit_part_label);
  gtk_entry_set_text (GTK_ENTRY (priv->label_entry),
                      gdu_device_partition_get_label (device));
*/
  /*  Volume label  */
  if (! priv->job_running) {
    gtk_entry_set_text (GTK_ENTRY (priv->label_entry),
                        gdu_presentable_get_name (priv->presentable));
  }


  /*  update the partition type combo  */
  if (! priv->job_running) {
    type_combo_box_changed (priv->part_type_combo_box, priv);
  }

  if (device != NULL)
    g_object_unref (device);

  if (toplevel_presentable != NULL)
    g_object_unref (toplevel_presentable);

  if (toplevel_device != NULL)
    g_object_unref (toplevel_device);
}


static void
nautilus_gdu_show_error (GtkWidget      *parent_window,
                         GduPresentable *presentable,
                         GError         *error,
                         const char     *primary_markup_format,
                         ...)
{
  GtkWidget *dialog;
  char *error_text;
  char *window_title;
  char *window_icon_name;
  va_list args;

  g_return_if_fail (presentable != NULL);
  g_return_if_fail (error != NULL);

  /* TODO: this still needs work */

  window_title = gdu_presentable_get_name (presentable);
  window_icon_name = gdu_presentable_get_icon_name (presentable);

  va_start (args, primary_markup_format);
  error_text = g_strdup_vprintf (primary_markup_format, args);
  va_end (args);

  dialog = gtk_message_dialog_new_with_markup (
          GTK_WINDOW (parent_window),
          GTK_DIALOG_MODAL|GTK_DIALOG_DESTROY_WITH_PARENT,
          GTK_MESSAGE_ERROR,
          GTK_BUTTONS_CLOSE,
          "<big><b>%s</b></big>",
          error_text);
  gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog), "%s", error->message);

  gtk_window_set_title (GTK_WINDOW (dialog), window_title);
  gtk_window_set_icon_name (GTK_WINDOW (dialog), window_icon_name);

  g_signal_connect_swapped (dialog,
                            "response",
                            G_CALLBACK (gtk_widget_destroy),
                            dialog);
  gtk_window_present (GTK_WINDOW (dialog));

  g_free (window_title);
  g_free (window_icon_name);
  g_free (error_text);
}

static void
free_format_action_data (FormatProcessData *data)
{
  if (data) {
    update_ui_progress (data->priv, data, FALSE);
    if (data->device != NULL)
      g_object_unref (data->device);
    if (data->encrypt_passphrase != NULL)
      g_free (data->encrypt_passphrase);
    if (data->fstype != NULL)
      g_free (data->fstype);
    if (data->fslabel != NULL)
      g_free (data->fslabel);
    if (data->secure_erase != NULL)
      g_free (data->secure_erase);
    g_free (data);
  }
}

/* ---------------------------------------------------------------------------------------------------- */

static void
modify_partition_completed (GduDevice  *device,
                            GError     *error,
                            gpointer    user_data)
{
  FormatProcessData *data = user_data;

  g_print ("modify_partition_completed\n");
  g_return_if_fail (data != NULL);

  if (error != NULL) {
    nautilus_gdu_show_error (GTK_WIDGET (data->priv->dialog),
                             data->priv->presentable,
                             error,
                             _("Error modifying partition"));
    g_error_free (error);
  }
  else
  {
    /*  -- don't refresh here, wait for the "changed" callback
    update_ui (data->priv);  */
  }
  free_format_action_data (data);
}

static void
part_modify_action_callback (GtkAction *action, gpointer user_data)
{
  FormatProcessData *data = user_data;

  g_return_if_fail (data != NULL);
  g_print ("part_modify_action_callback\n");

  if (data->priv->job_cancelled)
    return;

  /*  DK is buggy, passing a label string causes the operation to fail  */
  gdu_device_op_partition_modify (data->device, data->recommended_part_type, NULL /* data->fslabel */, NULL, modify_partition_completed, data);
}

static void
part_modify_auth_end_callback (PolKitGnomeAction *action,
                          gboolean           gained_privilege,
                          gpointer           user_data)
{
  FormatProcessData *data = user_data;

  g_return_if_fail (data != NULL);
  g_print ("part_modify_auth_end_callback\n");

  if (! gained_privilege) {
    /*  cancel the whole operation  */
    free_format_action_data (data);
  }
  else {
    /*  positive reply should be handled by part_modify_action_callback  */
  }
}

/* ---------------------------------------------------------------------------------------------------- */

static void
format_action_completed (GduDevice  *device,
                         GError     *error,
                         gpointer    user_data)
{
  FormatProcessData *data = user_data;
  const char *part_type;
  const char *scheme;

  g_print ("format_action_completed\n");
  g_return_if_fail (data != NULL);

  if (error != NULL) {
    nautilus_gdu_show_error (GTK_WIDGET (data->priv->dialog),
                             data->priv->presentable,
                             error,
                             _("Error creating partition"));
    g_error_free (error);
  }
  else
  {
    /*  formatting finished  */
    /*  TODO: show encryption info somewhere?  */
    if (data->encrypt_passphrase != NULL) {
      /* now set the passphrase if requested */
      if (data->save_in_keyring || data->save_in_keyring_session) {
        gdu_util_save_secret (device,
                              data->encrypt_passphrase,
                              data->save_in_keyring_session);
      }
    }

    scheme = gdu_device_partition_get_scheme (device);
//    g_print ("format_action_completed: scheme = %s\n", scheme);
    /*  TODO: extend to other partition types once this feature is working in DK  */
    if (data->priv->job_cancelled && strcmp (scheme, "mbr") == 0) {
      part_type = gdu_device_partition_get_type (device);
      data->recommended_part_type = get_part_type_for_fs (data->fstype);
      g_print ("format_action_completed: part_type = %s, recommended_part_type = %s\n", part_type, data->recommended_part_type);
      /*  Change partition type if necessary  */
      if (strcasecmp (part_type, data->recommended_part_type) != 0 &&
          ! (mbr_part_is_linux_type (part_type) && mbr_part_is_linux_type (data->recommended_part_type)))
        {
          g_print ("changing part type to %s, device = %s\n", data->recommended_part_type, gdu_device_get_device_file (device));
          gtk_action_activate (GTK_ACTION (data->priv->part_modify_action));
          return;  /*  don't change the UI yet  */
        }
    }

    /*  -- don't refresh here, wait for the "changed" callback
    update_ui (data->priv);  */
  }
  free_format_action_data (data);
}

static void
format_action_callback (GtkAction *action, gpointer user_data)
{
  FormatProcessData *data = user_data;

  g_return_if_fail (data != NULL);
  g_print ("format_action_callback\n");

  if (data->priv->job_cancelled)
    return;

  gdu_device_op_filesystem_create (data->device,
                                   data->fstype,
                                   data->fslabel,
                                   data->secure_erase,
                                   data->encrypt_passphrase,
                                   data->take_ownership,
                                   format_action_completed,
                                   data);
}

static void
format_auth_end_callback (PolKitGnomeAction *action,
                          gboolean           gained_privilege,
                          gpointer           user_data)
{
  FormatProcessData *data = user_data;

  g_return_if_fail (data != NULL);
  g_print ("format_auth_end_callback\n");

  if (! gained_privilege) {
    /*  cancel the whole operation  */
    free_format_action_data (data);
  }
  else {
    /*  positive reply should be handled by format_action_callback  */
  }
}


/* ---------------------------------------------------------------------------------------------------- */

static gboolean
unmount_show_busy (gpointer user_data)
{
  FormatProcessData *data = user_data;

  if (gdu_util_dialog_show_filesystem_busy (GTK_WIDGET (data->priv->dialog), data->priv->presentable)) {
    /*  user managed to kill all applications; try again  */
    if (! data->priv->job_cancelled) {
      gtk_action_activate (GTK_ACTION (data->priv->unmount_action));
    }
  }
  else {
    /*  cancel the whole operation  */
    free_format_action_data (data);
  }

  return FALSE;
}

static void
unmount_action_completed (GduDevice  *device,
                          GError     *error,
                          gpointer    user_data)
{
  FormatProcessData *data = user_data;

  g_return_if_fail (data != NULL);

  g_print ("unmount_action_completed\n");
  update_ui_mounted (data->priv);

  if (error != NULL) {
    if (g_error_matches (error, GDU_ERROR, GDU_ERROR_BUSY)) {
      /*  show dialog in idle  */
      g_idle_add (unmount_show_busy, data);
    }
    else
    {
      nautilus_gdu_show_error (GTK_WIDGET (data->priv->dialog),
                               data->priv->presentable,
                               error,
                               _("Error unmounting device"));
      free_format_action_data (data);
    }
    g_error_free (error);
  }
  else
  {
    /*  TODO: maybe perform a refresh?  */
    g_print ("  formatting...\n");
    if (data->priv->job_cancelled)
      return;
    gtk_action_activate (GTK_ACTION (data->priv->format_action));
  }
}

static void
unmount_action_callback (GtkAction *action, gpointer user_data)
{
  FormatProcessData *data = user_data;

  g_return_if_fail (data != NULL);
  g_print ("unmount_action_callback\n");

  if (data->priv->job_cancelled)
    return;

  gdu_device_op_filesystem_unmount (data->device, unmount_action_completed, data);
}

static void
unmount_auth_end_callback (PolKitGnomeAction *action,
                           gboolean           gained_privilege,
                           gpointer           user_data)
{
  FormatProcessData *data = user_data;

  g_return_if_fail (data != NULL);
  g_print ("unmount_auth_end_callback\n");

  if (! gained_privilege) {
    /*  cancel the whole operation  */
    free_format_action_data (data);
  }
  else {
    /*  positive reply should be handled by unmount_action_callback  */
  }
}

/* ---------------------------------------------------------------------------------------------------- */

/*  taken from palimpsest/gdu-section-unrecognized.c  */
static void
do_format (FormatDialogPrivate *priv)
{
  FormatProcessData *data;
  GduPresentable *toplevel_presentable;
  GduDevice *toplevel_device;
  char *primary;
  char *secondary;
  char *drive_name;
  GduKnownFilesystem *kfs;


  priv->job_cancelled = FALSE;
  data = g_new0 (FormatProcessData, 1);
  data->priv = priv;
  primary = NULL;
  secondary = NULL;
  toplevel_presentable = NULL;
  toplevel_device = NULL;
  drive_name = NULL;

  data->device = gdu_presentable_get_device (priv->presentable);
  if (data->device == NULL) {
          g_warning ("%s: device is not supposed to be NULL",  __FUNCTION__);
          free_format_action_data (data);
          goto out;
  }

  toplevel_presentable = gdu_presentable_get_toplevel (priv->presentable);
  if (toplevel_presentable == NULL) {
          g_warning ("%s: no toplevel presentable",  __FUNCTION__);
  }
  toplevel_device = gdu_presentable_get_device (toplevel_presentable);
  if (toplevel_device == NULL) {
          g_warning ("%s: no device for toplevel presentable",  __FUNCTION__);
          free_format_action_data (data);
          goto out;
  }

  drive_name = gdu_presentable_get_name (toplevel_presentable);
  data->fstype = gdu_util_fstype_combo_box_get_selected (priv->part_type_combo_box);
  data->fslabel = g_strdup (GTK_WIDGET_IS_SENSITIVE (priv->label_entry) ?
                            gtk_entry_get_text (GTK_ENTRY (priv->label_entry)) : "");

  data->take_ownership = FALSE;
  kfs = gdu_pool_get_known_filesystem_by_id (priv->pool, data->fstype);
  if (kfs != NULL) {
          if (gdu_known_filesystem_get_supports_unix_owners (kfs) && gtk_toggle_button_get_active (
                      GTK_TOGGLE_BUTTON (priv->take_ownership_of_fs_check_button)))
                data->take_ownership = TRUE;
          g_object_unref (kfs);
  }

  primary = g_strdup (_("<b><big>Are you sure you want to create a new file system, deleting existing data?</big></b>"));

  if (gdu_device_is_partition (data->device)) {
          if (gdu_device_is_removable (toplevel_device)) {
                  secondary = g_strdup_printf (_("All data on partition %d on the media in \"%s\" will be "
                                                 "irrecovably erased. "
                                                 "Make sure important data is backed up. "
                                                 "This action cannot be undone."),
                                               gdu_device_partition_get_number (data->device),
                                               drive_name);
          } else {
                  secondary = g_strdup_printf (_("All data on partition %d of \"%s\" will be "
                                                 "irrecovably erased. "
                                                 "Make sure important data is backed up. "
                                                 "This action cannot be undone."),
                                               gdu_device_partition_get_number (data->device),
                                               drive_name);
          }
  } else {
          if (gdu_device_is_removable (toplevel_device)) {
                  secondary = g_strdup_printf (_("All data on the media in \"%s\" will be irrecovably erased. "
                                                 "Make sure important data is backed up. "
                                                 "This action cannot be undone."),
                                               drive_name);
          } else {
                  secondary = g_strdup_printf (_("All data on the drive \"%s\" will be irrecovably erased. "
                                                 "Make sure important data is backed up. "
                                                 "This action cannot be undone."),
                                               drive_name);
          }
  }


  update_ui_progress (priv, data, TRUE);

  data->secure_erase = gdu_util_delete_confirmation_dialog (GTK_WIDGET (priv->dialog),
                                                            "",
                                                            primary,
                                                            secondary,
                                                            _("C_reate"));

  if (data->secure_erase == NULL) {
    free_format_action_data (data);
    goto out;
  }

  if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (priv->encrypted_check_button))) {
    data->encrypt_passphrase = gdu_util_dialog_ask_for_new_secret (GTK_WIDGET (priv->dialog),
                                                                   &data->save_in_keyring,
                                                                   &data->save_in_keyring_session);
    if (data->encrypt_passphrase == NULL) {
      free_format_action_data (data);
      goto out;
    }
  }



  /*  unmount the device before formatting  */
  if (gdu_device_is_mounted (data->device)) {
    gtk_action_activate (GTK_ACTION (priv->unmount_action));
  } else {
    gtk_action_activate (GTK_ACTION (priv->format_action));
  }

out:
  if (toplevel_presentable != NULL)
    g_object_unref (toplevel_presentable);
  if (toplevel_device != NULL)
    g_object_unref (toplevel_device);
  g_free (primary);
  g_free (secondary);
  g_free (drive_name);
}


static void
nautilus_gdu_destroy (FormatDialogPrivate  *priv)
{
  g_return_if_fail (priv != NULL);
//  g_print ("...destroying...\n");

  /*  disconnect our handlers, since the presentable refence counter doesn't really need to be zero  */
  g_signal_handlers_disconnect_matched (priv->presentable, G_SIGNAL_MATCH_DATA, 0, 0, NULL, NULL, priv);

  /*  destroy PolicyKit actions  */
  polkit_action_unref (priv->pk_unmount_action);
  g_object_unref (priv->unmount_action);
  polkit_action_unref (priv->pk_format_action);
  g_object_unref (priv->format_action);
  polkit_action_unref (priv->pk_part_modify_action);
  g_object_unref (priv->part_modify_action);

  /*  destroy the dialog and internal struct  */
  gtk_widget_destroy (GTK_WIDGET (priv->dialog));
  g_object_unref (priv->pool);
  g_print ("before unreffing presentable, count = %d [%p]\n", ((GObject*)priv->presentable)->ref_count, priv->presentable);
  g_object_unref (priv->presentable);
  g_free (priv);
}

static void
presentable_removed (GduPresentable *presentable, FormatDialogPrivate *priv)
{
  g_return_if_fail (priv != NULL);
  g_warning ("Presentable removed event.\n");
  nautilus_gdu_destroy (priv);
}

static void
presentable_changed (GduPresentable *presentable, FormatDialogPrivate *priv)
{
  g_return_if_fail (priv != NULL);
  g_warning ("Presentable changed event.\n");

  /*  TODO: shall we preserve label or any other settings?  */
  update_ui (priv);
}

static void
cancel_operation (FormatDialogPrivate *priv)
{
  GduDevice *device;

  g_return_if_fail (priv != NULL);
  g_return_if_fail (priv->job_running == TRUE);
  g_return_if_fail (priv->presentable != NULL);
  g_warning ("Cancelling...\n");

  priv->job_cancelled = TRUE;
  device = gdu_presentable_get_device (priv->presentable);

  g_return_if_fail (device != NULL);
  gdu_device_op_cancel_job (device, NULL, NULL);
  g_object_unref (device);
}

static gboolean
window_delete_event (GtkWidget            *widget,
                     GdkEvent             *event,
                     FormatDialogPrivate  *priv)
{
  g_return_val_if_fail (priv != NULL, FALSE);
  if (priv->job_running) {
    cancel_operation (priv);
    return TRUE;  /*  consume the event  */
  }
  return FALSE;
}

static void
format_dialog_got_response (GtkDialog            *dialog,
                            gint                  response_id,
                            FormatDialogPrivate  *priv)
{
  if (response_id == GTK_RESPONSE_OK) {
    do_format (priv);
  }
  else
  if (response_id == GTK_RESPONSE_REJECT) {
    /*  revert changes  */
    update_ui (priv);
  }
  else {
    if (priv->job_running) {
      cancel_operation (priv);
    }
    else {
      /*  destroy the window and unref the presentable  */
      nautilus_gdu_destroy (priv);
    }
  }
}

void
nautilus_gdu_spawn_dialog (GduPresentable *presentable)
{
  GtkDialog *dialog;
  GtkWidget *content_area;
  FormatDialogPrivate *priv;


  g_return_if_fail (presentable != NULL);


  priv = g_new0 (FormatDialogPrivate, 1);
  priv->presentable = g_object_ref (presentable);
  priv->pool = gdu_presentable_get_pool (priv->presentable);
  priv->job_running = FALSE;

  g_print ("nautilus_gdu_spawn_dialog: after reffing presentable, count = %d [%p]\n", ((GObject*)priv->presentable)->ref_count, priv->presentable);

  dialog = GTK_DIALOG (gtk_dialog_new ());
  priv->dialog = dialog;

  /*  HIG stuff...  */
  gtk_dialog_set_has_separator (dialog, FALSE);
  gtk_container_set_border_width (GTK_CONTAINER (dialog), 5);
  gtk_box_set_spacing (GTK_BOX (dialog->vbox), 2);
  gtk_container_set_border_width (GTK_CONTAINER (dialog->action_area), 5);
  gtk_box_set_spacing (GTK_BOX (dialog->action_area), 6);

  gtk_window_set_resizable (GTK_WINDOW (dialog), FALSE);
  gtk_window_set_title (GTK_WINDOW (dialog), "");
  /*  TODO: change icon to something more decent  */
  gtk_window_set_icon_name (GTK_WINDOW (dialog), "palimpsest");

  gtk_dialog_add_buttons (dialog,
                          _("_Format"), GTK_RESPONSE_OK,
                          _("_Revert"), GTK_RESPONSE_REJECT,
                          NULL);
  priv->close_button = gtk_dialog_add_button (dialog, GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE);
  gtk_dialog_set_default_response (dialog, GTK_RESPONSE_CLOSE);

//  gtk_window_set_default_size (GTK_WINDOW (dialog), 300, 400);
  content_area = gtk_dialog_get_content_area (GTK_DIALOG (dialog));
  gtk_container_set_border_width (GTK_CONTAINER (content_area), 10);


  /*  headlines  */
  GtkWidget *label;
  GtkWidget *align;
  GtkWidget *vbox3;
  GtkWidget *vbox2;
  GtkWidget *hbox;
  GtkWidget *image;
  GtkWidget *table;
  GtkWidget *entry;
  GtkWidget *combo_box;
  GtkWidget *check_button;
  GtkWidget *progress_bar;
  int row;

  hbox = gtk_hbox_new (FALSE, 10);
  gtk_box_pack_start (GTK_BOX (content_area), hbox, FALSE, TRUE, 0);

  image = gtk_image_new ();
  gtk_box_pack_start (GTK_BOX (hbox), image, FALSE, TRUE, 0);
  priv->icon_image = image;

  vbox3 = gtk_vbox_new (FALSE, 0);
  align = gtk_alignment_new (0.0, 0.5, 0.0, 0.0);
  gtk_container_add (GTK_CONTAINER (align), vbox3);
  gtk_box_pack_start (GTK_BOX (hbox), align, FALSE, TRUE, 0);

  label = gtk_label_new (NULL);
  gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
  gtk_box_pack_start (GTK_BOX (vbox3), label, FALSE, TRUE, 0);
  priv->name_label = label;

  label = gtk_label_new (NULL);
  gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
  gtk_box_pack_start (GTK_BOX (vbox3), label, FALSE, TRUE, 0);
  priv->details1_label = label;

  label = gtk_label_new (NULL);
  gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
  gtk_box_pack_start (GTK_BOX (vbox3), label, FALSE, TRUE, 0);
  priv->details2_label = label;

  label = gtk_label_new (NULL);
  gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
  gtk_box_pack_start (GTK_BOX (vbox3), label, FALSE, TRUE, 0);
  priv->details3_label = label;



  /*  partition  */
  vbox3 = gtk_vbox_new (FALSE, 2);
  align = gtk_alignment_new (0.5, 0.5, 1.0, 1.0);
  gtk_alignment_set_padding (GTK_ALIGNMENT (align), 10, 30, 24, 0);
  gtk_container_add (GTK_CONTAINER (align), vbox3);
  gtk_box_pack_start (GTK_BOX (content_area), align, FALSE, TRUE, 0);

  label = gtk_label_new (NULL);
  gtk_label_set_markup (GTK_LABEL (label), _("<b>Format</b>"));
  gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
  gtk_box_pack_start (GTK_BOX (vbox3), label, FALSE, FALSE, 0);
  priv->partition_label = label;

  vbox2 = gtk_vbox_new (FALSE, 5);
  align = gtk_alignment_new (0.5, 0.5, 1.0, 1.0);
  gtk_alignment_set_padding (GTK_ALIGNMENT (align), 0, 0, 24, 0);
  gtk_container_add (GTK_CONTAINER (align), vbox2);
  gtk_box_pack_start (GTK_BOX (vbox3), align, FALSE, TRUE, 0);
  priv->controls_box = vbox2;

  /* explanatory text */
  label = gtk_label_new (NULL);
  gtk_label_set_markup (GTK_LABEL (label), _("To create a new file system on the device, select the type "
                                             "and label and then press \"Format\". All existing data will "
                                             "be lost."));
  gtk_label_set_line_wrap (GTK_LABEL (label), TRUE);
  gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
  gtk_box_pack_start (GTK_BOX (vbox2), label, FALSE, TRUE, 0);

  row = 0;

  table = gtk_table_new (2, 2, FALSE);
  gtk_box_pack_start (GTK_BOX (vbox2), table, FALSE, FALSE, 0);

  /*  partition label  */
  label = gtk_label_new (NULL);
  gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
  gtk_label_set_markup_with_mnemonic (GTK_LABEL (label), _("_Label:"));
  gtk_table_attach (GTK_TABLE (table), label, 0, 1, row, row + 1,
                    GTK_FILL, GTK_EXPAND | GTK_FILL, 2, 2);
  entry = gtk_entry_new ();
  gtk_table_attach (GTK_TABLE (table), entry, 1, 2, row, row + 1,
                    GTK_FILL, GTK_EXPAND | GTK_FILL, 2, 2);
  gtk_label_set_mnemonic_widget (GTK_LABEL (label), entry);
  priv->label_entry = entry;
  row++;

  /*  partition type  */
  label = gtk_label_new (NULL);
  gtk_misc_set_alignment (GTK_MISC (label), 1.0, 0.5);
  gtk_label_set_markup_with_mnemonic (GTK_LABEL (label), _("_Type:"));
  gtk_table_attach (GTK_TABLE (table), label, 0, 1, row, row + 1,
                    GTK_FILL, GTK_EXPAND | GTK_FILL, 2, 2);
  combo_box = gdu_util_fstype_combo_box_create (priv->pool, NULL);
  gtk_table_attach (GTK_TABLE (table), combo_box, 1, 2, row, row +1,
                    GTK_FILL, GTK_EXPAND | GTK_FILL, 2, 2);
  gtk_label_set_mnemonic_widget (GTK_LABEL (label), combo_box);
  priv->part_type_combo_box = combo_box;
  row++;

  /* type desc */
  label = gtk_label_new (NULL);
  gtk_label_set_line_wrap (GTK_LABEL (label), TRUE);
  gtk_label_set_width_chars (GTK_LABEL (label), 40);
  gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
  gtk_table_attach (GTK_TABLE (table), label, 1, 2, row, row + 1,
                    GTK_FILL, GTK_EXPAND | GTK_FILL, 2, 2);
  gdu_util_fstype_combo_box_set_desc_label (combo_box, label);
  row++;

  /*  flags: used by mbr, apm  */
  check_button = gtk_check_button_new_with_mnemonic (_("_Bootable"));
  gtk_table_attach (GTK_TABLE (table), check_button, 1, 2, row, row +1,
                    GTK_FILL, GTK_EXPAND | GTK_FILL, 2, 2);
  priv->part_flag_boot_check_button = check_button;
  row++;

  /*  flags: used by gpt  */
  check_button = gtk_check_button_new_with_mnemonic (_("Required / Firm_ware"));
  gtk_table_attach (GTK_TABLE (table), check_button, 1, 2, row, row +1,
                    GTK_FILL, GTK_EXPAND | GTK_FILL, 2, 2);
  priv->part_flag_required_check_button = check_button;
  row++;

  /* whether to chown fs root for user */
  check_button = gtk_check_button_new_with_mnemonic (_("T_ake ownership of file system"));
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (check_button), TRUE);
  gtk_widget_set_tooltip_text (check_button,
                               _("The selected file system has a concept of file ownership. "
                                 "If checked, the created file system be will be owned by you. "
                                 "If not checked, only the super user can access the file system."));
  gtk_table_attach (GTK_TABLE (table), check_button, 1, 2, row, row + 1,
                    GTK_FILL, GTK_EXPAND | GTK_FILL, 2, 2);
  priv->take_ownership_of_fs_check_button = check_button;
  row++;

  /* whether to encrypt underlying device */
  check_button = gtk_check_button_new_with_mnemonic (_("E_ncrypt underlying device"));
  gtk_widget_set_tooltip_text (check_button,
                               _("Encryption protects your data, requiring a "
                                 "passphrase to be enterered before the file system can be "
                                 "used. May decrease performance and may not be compatible if "
                                 "you use the media on other operating systems."));
  gtk_table_attach (GTK_TABLE (table), check_button, 1, 2, row, row + 1,
                    GTK_FILL, GTK_EXPAND | GTK_FILL, 2, 2);
  priv->encrypted_check_button = check_button;
  row++;

  /*  progress bar  */
  align = gtk_alignment_new (0.5, 0.5, 1.0, 1.0);
  gtk_alignment_set_padding (GTK_ALIGNMENT (align), 15, 0, 20, 25);
  priv->progress_bar_box = align;

  progress_bar = gtk_progress_bar_new ();
  priv->progress_bar = progress_bar;
  gtk_container_add (GTK_CONTAINER (align), progress_bar);
  gtk_box_pack_start (GTK_BOX (vbox3), align, TRUE, TRUE, 0);


  /*  mounted warning box  */
  hbox = gtk_hbox_new (FALSE, 7);
  align = gtk_alignment_new (0.5, 0.5, 1.0, 1.0);
  gtk_alignment_set_padding (GTK_ALIGNMENT (align), 15, 15, 20, 0);
  gtk_container_add (GTK_CONTAINER (align), hbox);
  gtk_box_pack_start (GTK_BOX (content_area), align, FALSE, TRUE, 0);
  priv->mount_warning = align;

  image = gtk_image_new_from_stock (GTK_STOCK_DIALOG_WARNING, GTK_ICON_SIZE_LARGE_TOOLBAR);
  gtk_box_pack_start (GTK_BOX (hbox), image, FALSE, TRUE, 0);

  label = gtk_label_new (_("The volume is currently mounted. Please make sure to close all open files before formatting."));
//  gtk_label_set_width_chars (GTK_LABEL (label), 50);
  gtk_label_set_line_wrap (GTK_LABEL (label), TRUE);
  gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
  gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, TRUE, 0);


  /*  set up PolicyKit actions  */
  priv->pk_unmount_action = polkit_action_new ();
  polkit_action_set_action_id (priv->pk_unmount_action, "org.freedesktop.devicekit.disks.filesystem-unmount-others");
  priv->unmount_action = polkit_gnome_action_new_default ("unmount", priv->pk_unmount_action, NULL, NULL);
  priv->pk_format_action = polkit_action_new ();
  polkit_action_set_action_id (priv->pk_format_action, "org.freedesktop.devicekit.disks.change");
  priv->format_action = polkit_gnome_action_new_default ("format", priv->pk_format_action, NULL, NULL);
  priv->pk_part_modify_action = polkit_action_new ();
  /*  action_id is the same as for format, but sometimes authentication is one shot  */
  polkit_action_set_action_id (priv->pk_part_modify_action, "org.freedesktop.devicekit.disks.change");
  priv->part_modify_action = polkit_gnome_action_new_default ("part_modify", priv->pk_part_modify_action, NULL, NULL);


  //  gtk_dialog_set_response_sensitive (dialog, GTK_RESPONSE_OK, FALSE);

  g_signal_connect (priv->dialog, "delete-event",
                    G_CALLBACK (window_delete_event), priv);
  /*  update sensivity and length of fs label + entry  */
  g_signal_connect (priv->part_type_combo_box, "changed",
                    G_CALLBACK (type_combo_box_changed), priv);
  g_signal_connect (G_OBJECT (dialog), "response",
                    G_CALLBACK (format_dialog_got_response), priv);
  /*  catch Presentable events  */
  g_signal_connect (G_OBJECT (priv->presentable), "removed",
                    G_CALLBACK (presentable_removed), priv);
  g_signal_connect (G_OBJECT (priv->presentable), "changed",
                    G_CALLBACK (presentable_changed), priv);

  gtk_widget_show_all (GTK_WIDGET (dialog));
  gtk_widget_grab_focus (priv->close_button);
  update_ui (priv);
  update_ui_progress (priv, NULL, FALSE);
}
